import java.awt.Color;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.InputMismatchException;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;
import java.util.TreeMap;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.json.JSONObject;

public class NonogramHelper extends JFrame implements ActionListener {
	private static final String APPID = "742500";
	
	private Map<String, SaveData> incompleteSaves = null;
	private Map<String, WorkshopData> workshopMap = null;
	
	private JComboBox<WorkshopData> puzzleSelectCombobox = null;
	private JScrollPane currentProgressPane = new JScrollPane();
	private JScrollPane errorsPane = new JScrollPane();
	private JScrollPane solutionPane = new JScrollPane();
	private JScrollPane imagePane = new JScrollPane();
	
	public NonogramHelper() {
		// preinit data before doing anything
		initialize();
		
		Container pane = this.getContentPane();
		((JPanel) pane).setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		JPanel mainPanel = new JPanel();
		mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
		
		JPanel selectBoxPanel = new JPanel();
		selectBoxPanel.setLayout(new BoxLayout(selectBoxPanel, BoxLayout.X_AXIS));
		selectBoxPanel.add(new JLabel("Select the puzzle that gives you issues: "));
		selectBoxPanel.add(Box.createHorizontalStrut(10));
		WorkshopData[] data = workshopMap.values().toArray(new WorkshopData[0]);
		puzzleSelectCombobox = new JComboBox<>(data);
		puzzleSelectCombobox.setMaximumSize(puzzleSelectCombobox.getPreferredSize());
		puzzleSelectCombobox.setSelectedIndex(-1); // no selection despite no empty entry
		puzzleSelectCombobox.addActionListener(this);
		selectBoxPanel.add(puzzleSelectCombobox);
		selectBoxPanel.add(Box.createHorizontalGlue());
		
		mainPanel.add(selectBoxPanel);
		mainPanel.add(Box.createVerticalStrut(10));
		
		JTabbedPane tabbedPane = new JTabbedPane();
		tabbedPane.addTab("Current Progress (Spoiler free)", currentProgressPane);
		tabbedPane.addTab("Mistakes marked", errorsPane);
		tabbedPane.addTab("Completed solution", solutionPane);
		tabbedPane.addTab("Result image", imagePane);
		tabbedPane.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
		tabbedPane.setPreferredSize(tabbedPane.getMaximumSize());
		mainPanel.add(tabbedPane);
		
		this.add(mainPanel);
		
		this.setSize(800, 800);
		this.setLocationRelativeTo(null);
		this.setVisible(true);
		this.setTitle("Akhlys little Nonogram Master's Legacy helper");
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	}
	
	private void initialize() {
		// steam path
		JFileChooser jfc = new JFileChooser();
		Path steamPath = SteamDirectoryHelper.getSteamInstallPath();
		if (steamPath == null) {
			JOptionPane.showMessageDialog(null, "Couldn't find Steam install dir. Please select it manually.\nDefault location is C:\\Program Files (x86)\\Steam", "Steam not found", JOptionPane.ERROR_MESSAGE);
			jfc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
			jfc.setDialogTitle("Select steam directory");
			int option = jfc.showOpenDialog(null);
			if (option == JFileChooser.APPROVE_OPTION) {
				steamPath = Paths.get(jfc.getSelectedFile().getAbsolutePath());
				if (!Files.exists(steamPath.resolve("steamapps"))) {
					JOptionPane.showMessageDialog(null, "Doesn't look like a Steam directory (no steamapps), will exit now", "Steam not found", JOptionPane.ERROR_MESSAGE);
					System.exit(0);
				}
			} else {
				System.exit(0);
			}
		}
		
		// saved games
		Path saveGame = null;
		Map<String, Path> saveGames = null;
		String promptMsg = "Please select it manually.\nDefault location is C:\\Program Files (x86)\\Steam\\userdata\\<user-id>\\" + APPID + "\\remote\\saves.dat";
		try {
			saveGames = getSaveGames(steamPath, APPID);
			if (saveGames.size() > 1) {
				String[] choices = saveGames.keySet().toArray(new String[0]);
				String input = (String) JOptionPane.showInputDialog(null, "Multiple Steam-Users with save data found, please select the Steam user ID that you want",
				        "Select the correct Steam user ID", JOptionPane.QUESTION_MESSAGE, null, choices, choices[1]);
				if (input == null) System.exit(0);
				saveGame = saveGames.get(input);
			} else if (saveGames.isEmpty()) {
				JOptionPane.showMessageDialog(null, "Couldn't find any Nonogram Master's Legacy save games\n" + promptMsg, "Savegame not found", JOptionPane.ERROR_MESSAGE);
			} else {
				saveGame = saveGames.values().iterator().next();
			}
		} catch (IOException e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(null, "Couldn't find any Nonogram Master's Legacy save games due to error: " + e.getMessage() + "\n" + promptMsg, "Savegame not found", JOptionPane.ERROR_MESSAGE);
		}
		if (saveGame == null) {
			jfc.setFileSelectionMode(JFileChooser.FILES_ONLY);
			jfc.setFileFilter(new FileNameExtensionFilter("save file","dat"));
			jfc.setDialogTitle("Select Nonogram Master's Legacy save file");
			int option = jfc.showOpenDialog(null);
			if (option == JFileChooser.APPROVE_OPTION) {
				saveGame = Paths.get(jfc.getSelectedFile().getAbsolutePath());
			} else {
				System.exit(0);
			}
		}
		
		// workshop
		Path workshopDir = null;
		try {
			workshopDir = SteamDirectoryHelper.findWorkshopPath(steamPath, APPID);
		} catch (IOException e) {
			e.printStackTrace();
		}
		if (workshopDir == null) {
			JOptionPane.showMessageDialog(null, "Couldn't find Nonogram Master's Legacy's Workshop directory, please select it manually\nDefault location is C:\\Program Files (x86)\\Steam\\SteamApps\\workshop\\content\\" + APPID, "Workshop not found", JOptionPane.ERROR_MESSAGE);
			jfc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
			jfc.setDialogTitle("Select workshop directory");
			int option = jfc.showOpenDialog(null);
			if (option == JFileChooser.APPROVE_OPTION) {
				workshopDir = Paths.get(jfc.getSelectedFile().getAbsolutePath());
			} else {
				System.exit(0);
			}
		}
		
		try {
			String saveContent = new String(Files.readAllBytes(saveGame), StandardCharsets.UTF_8);
			JSONObject saveObj = new JSONObject(saveContent);
			this.incompleteSaves = getIncompleteGames(saveObj);
			
			if (this.incompleteSaves.size() == 0) {
				JOptionPane.showMessageDialog(null, "You don't seem to have any puzzles that are in progress. I'm afraid this app is not meant for completed ones.\nIf you have the puzzle currently open, try \"Back\" in-game to save your progress and try again", "No in progress games", JOptionPane.INFORMATION_MESSAGE);
				System.exit(0);
			}
			
			this.workshopMap = getWorkshopItemData(workshopDir, this.incompleteSaves.keySet());
		} catch (IOException e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(null, "Couldn't load save data or workshop data, will exit: " + e.getMessage(), "Input/Output error", JOptionPane.ERROR_MESSAGE);
		}
		
	}
	
	
	public static void main(String[] args) throws IOException {
		new NonogramHelper();
	}
	
	private static Map<String, SaveData> getIncompleteGames(JSONObject save) {
		// for some reason, levelstates are json that is encoded as a string and attached to this json
		JSONObject workShopLevelStates = new JSONObject(save.getString("PP_WORKSHOP_LEVEL_STATES"));
		Iterator<String> keys = workShopLevelStates.keys();
		Map<String, SaveData> incompleteLevels = new TreeMap<>();
		while(keys.hasNext()) {
			String key = keys.next();
			JSONObject gameState = workShopLevelStates.getJSONObject(key);
			if (!gameState.getBoolean("inProgress")) continue;
			incompleteLevels.put(key, new SaveData(gameState));
		}
		return incompleteLevels;
	}
	
	private static Map<String, WorkshopData> getWorkshopItemData(Path workshopDir, Set<String> workshopIds) throws IOException {
		Map<String, WorkshopData> data = new TreeMap<>();
		for (String id : workshopIds) {
			Path dir = workshopDir.resolve(id);
			String levelContent = new String(Files.readAllBytes(dir.resolve("Level.bytes")), StandardCharsets.UTF_8);
			data.put(id, new WorkshopData(id, new JSONObject(levelContent)));
		}
		
		return data;
	}
	
	private static Map<String, Path> getSaveGames(Path steamInstallPath, String gameId) throws IOException {
		//Paths.get("C:\\Program Files (x86)\\Steam\\userdata\\83206599\\742500\\remote\\saves.dat");
		Path userData = steamInstallPath.resolve("userdata");
		Map<String, Path> saveGames = new TreeMap<>(); 
		try (DirectoryStream<Path> stream = Files.newDirectoryStream(userData)) {
	        for (Path path : stream) {
	            if (!Files.isDirectory(path)) continue;
	            Path saveGame = path.resolve(gameId + "/remote/saves.dat");
	            if (Files.exists(saveGame)) {
	            	saveGames.put(path.getFileName().toString(), saveGame);
	            }
	        }
	    }
		return saveGames;
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		WorkshopData puzzle = (WorkshopData)puzzleSelectCombobox.getSelectedItem();
		if (puzzle == null) return;
		SaveData saveData = this.incompleteSaves.get(puzzle.getWorkshopID());
		saveData.initDimensions(puzzle.getWidth(), puzzle.getHeight());
		
		BufferedImage progress = saveData.getCurrentProgressImageGrid();
		BufferedImage errors = saveData.getErrorImageGrid(puzzle.getPuzzleData());
		BufferedImage solution = puzzle.getSolutionImageGrid();
		BufferedImage complete = puzzle.getClearImage(10);

		currentProgressPane.setViewportView(new JLabel(new ImageIcon(progress)));
		errorsPane.setViewportView(new JLabel(new ImageIcon(errors)));
		solutionPane.setViewportView(new JLabel(new ImageIcon(solution)));
		imagePane.setViewportView(new JLabel(new ImageIcon(complete)));
	}
}
